import pycomm3

from . import helper
from . import registry
from .. import types

# Known RemoteHelper file version numbers, used to help check that device is a valid terminal.
HELPER_VERSIONS = {
    '5.10.00',
    '6.00.00',
    '6.10.00',
    '7.00.00',
    '8.00.00',
    '8.10.00',
    '8.20.00',
    '9.00.00',
    '10.00.00',
    '11.00.00',
    '12.00.00',
    '12.00.00.414.56',
    '12.00.00.414.73',
    '12.00.00.414.85',
    '12.00.00.414.99',
    '12.00.00.414.106',
    '13.00.00',
    '14.00.00',
    '15.00.00'
}

# Known terminal MEVersion numbers, used to help check that device is a valid terminal.
ME_VERSIONS = {
    '5.10.16.09',
    '6.00.04.16',
    '6.10.17.09',
    '7.00.20.13',
    '7.00.55.13',
    '8.00.67.12',
    '8.10.42.13',
    '8.20.30.10',
    '9.00.17.241',
    '9.00.00.241',
    '10.00.09.290',
    '11.00.00.230',
    '11.00.25.230',
    '12.00.00.414',
    '12.00.00.414.56.47',
    '12.00.00.414.85.414',
    '12.00.00.414.99.414',
    '12.00.00.414.106.414',
    '12.00.77.414',
    '12.00.78.414',
    '13.00.11.413',
    '14.00.00.394',
    '15.00.00.201'
}

# Known terminal product codes, used to help check that device is a valid terminal.
PRODUCT_CODES = {
    13,   #PanelView Plus 700 CE
    14,   #PanelView Plus 700
    15,   #PanelView Plus 400
    16,   #PanelView Plus 600
    17,   #PanelView Plus 1000
    18,   #PanelView Plus 1250
    19,   #PanelView Plus 1500
    20,   #PanelView Plus 1000 CE
    21,   #PanelView Plus 1250 CE
    22,   #PanelView Plus 1500 CE
    40,   #PanelView Plus Compact 400
    42,   #PanelView Plus Compact 600
    44,   #PanelView Plus Compact 1000
    45,   #PanelView Plus 6 700
    46,   #PanelView Plus 6 700 CE
    47,   #PanelView Plus 6 1000
    48,   #PanelView Plus 6 1000 CE
    49,   #PanelView Plus 6 1250
    50,   #PanelView Plus 6 1250 CE
    51,   #PanelView Plus 6 1500
    52,   #PanelView Plus 6 1500 CE
    74,   #PanelView Plus 6 400
    75,   #PanelView Plus 6 400
    77,   #PanelView Plus 6 600
    78,   #PanelView Plus 6 600 Extended
    79,   #PanelView Plus 6 600 Extended
    81,   #PanelView Plus 7 Standard 600
    83,   #PanelView Plus 7 Standard 700
    85,   #PanelView Plus 7 Standard 900W
    87,   #PanelView Plus 7 Standard 1000
    89,   #PanelView Plus 7 Standard 1200W
    91,   #PanelView Plus 7 Standard 1500
    94,   #PanelView Plus 7 Performance 700
    96,   #PanelView Plus 7 Performance 900W
    98,   #PanelView Plus 7 Performance 1000
    100,  #PanelView Plus 7 Performance 1200W
    102,  #PanelView Plus 7 Performance 1500
    104,  #PanelView Plus 7 Performance 1900
    #107, #PanelView Plus 7 Performance Series B DLR Network Switch - not a terminal
    110,  #PanelView Plus 7 Performance 700W
    112,  #PanelView Plus 7 Performance 1000W
    114,  #PanelView Plus 7 Performance 1200W
    116,  #PanelView Plus 7 Performance 1500W
    118,  #PanelView Plus 7 Performance 1900W
    147,  #PanelView Plus 6 Compact 1000
    175,  #PanelView Plus 6 Compact 400
    177,  #PanelView Plus 6 Compact 600
    179,  #PanelView Plus 7 Standard 400W
    180,  #PanelView Plus 7 Standard 400W DLR
    181,  #PanelView Plus 7 Standard 600
    182,  #PanelView Plus 7 Standard 600 DLR
    183,  #PanelView Plus 7 Standard 700
    184,  #PanelView Plus 7 Standard 700 DLR
    185,  #PanelView Plus 7 Standard 900W
    186,  #PanelView Plus 7 Standard 900W DLR
    187,  #PanelView Plus 7 Standard 1000
    188,  #PanelView Plus 7 Standard 1000 DLR
    189,  #PanelView Plus 7 Standard 1200W
    190,  #PanelView Plus 7 Standard 1200W DLR
    191,  #PanelView Plus 7 Standard 1500
    192   #PanelView Plus 7 Standard 1500 DLR
}

# Known product types, used to help check that device is a valid terminal.
PRODUCT_TYPES = {
    24
}

HELPER_FILE_NAME = 'RemoteHelper.DLL'
UPLOAD_LIST_PATH = 'Rockwell Software\\RSViewME\\Runtime\\Results.txt'

def get_terminal_info(cip: pycomm3.CIPDriver) -> types.MEDeviceInfo:
    me_version = registry.get_me_version(cip)
    major_rev = int(me_version.split(".")[0])

    if major_rev <= 5:
        helper_path = '\\Storage Card\\Rockwell Software\\RSViewME'
        storage_path = '\\Storage Card'
    else:
        helper_path = '\\Windows'
        storage_path = '\\Application Data'

    helper_file_path = f'{helper_path}\\{HELPER_FILE_NAME}'
    upload_list_path = f'{storage_path}\\{UPLOAD_LIST_PATH}'
    paths = types.MEDevicePaths(helper_file_path,
                              storage_path,
                              upload_list_path)

    return types.MEDeviceInfo(cip._cip_path, 
                               helper.get_helper_version(cip, paths),
                               me_version,
                               registry.get_version_major(cip),
                               registry.get_version_minor(cip),
                               registry.get_product_code(cip),
                               registry.get_product_name(cip),
                               registry.get_product_type(cip),
                               [],
                               [],
                               '',
                               '',
                               paths)

def extract_version_prefix(version: str) -> str:
    """Extracts the major and minor version (e.g., '12.00') from a version string."""
    return '.'.join(version.split('.')[:2])

def is_version_matched(device_version: str, known_versions: set) -> bool:
    """Checks if the device version prefix matches any of the known version prefixes."""
    device_version_prefix = extract_version_prefix(device_version)
    return any(extract_version_prefix(known_version) == device_version_prefix for known_version in known_versions)

def is_terminal_valid(device: types.MEDeviceInfo) -> bool:
    if device.product_type not in PRODUCT_TYPES: return False
    if device.product_code not in PRODUCT_CODES: return False
    if not is_version_matched(device.helper_version, HELPER_VERSIONS): return False
    if not is_version_matched(device.me_version, ME_VERSIONS): return False
    return True

def is_download_valid(cip: pycomm3.CIPDriver, device: types.MEDeviceInfo, file: types.MEFile) -> bool:
    # Check that file is correct extension
    if (file.get_ext() != '.mer'):
        device.log.append(f'File {file.name} is not a *.mer file')
        return False

    # Check that storage folder exists
    resp_storage_exists = helper.get_folder_exists(cip, device.paths)
    if not(resp_storage_exists):
        device.log.append(f'Storage folder does not exist on terminal')
        return False

    # Check free space
    resp_free_space = helper.get_free_space(cip, device.paths)
    if (resp_free_space > file.get_size()):
        device.log.append(f'File {file.name} requires {file.get_size()} byes.  Free space on terminal {resp_free_space} bytes.')
    else:
        device.log.append(f'File {file.name} requires {file.get_size()} bytes.  Free space on terminal {resp_free_space} bytes is insufficient.')
        return False

    # Check if file name already exists
    resp_file_exists = helper.get_file_exists(cip, device.paths, file)
    file.overwrite_required = False
    if (resp_file_exists and file.overwrite_requested):
        device.log.append(f'File {file.name} already exists on terminal, and overwrite was requested.  Setting overwrite to required.')
        file.overwrite_required = True
    if (resp_file_exists and not(file.overwrite_requested)):
        device.log.append(f'File {file.name} already exists on terminal, and overwrite was NOT requested.  Use kwarg overwrite_requested=True to overwrite existing.')
        return False
    if not(resp_file_exists):
        device.log.append(f'File {file.name} does not exist on terminal.  Setting overwrite to not required.')

    # Check space consumed by file if it exists
    if resp_file_exists:
        resp_file_size = helper.get_file_size(cip, device.paths, file)
        device.log.append(f'File {file.name} on terminal is {resp_file_size} bytes.')

    return True


import os
import pycomm3

from . import files
from . import helper
from . import registry
from .. import types

def create_log(cip: pycomm3.CIPDriver, device: types.MEDeviceInfo, print_log: bool, redact_log: bool, silent_mode: bool):
    if print_log: print(f'Terminal product type: {device.product_type}.')
    if print_log: print(f'Terminal product code: {device.product_code}.')
    if print_log: print(f'Terminal product name: {device.product_name}.')
    if print_log: print(f'Terminal helper version: {device.helper_version}.')
    if print_log: print(f'Terminal ME version: {device.me_version}.')
    if print_log: print(f'Terminal major version: {device.version_major}.')
    if print_log: print(f'Terminal minor version: {device.version_minor}.')
    if print_log: print(f'Terminal helper path: {device.paths.helper_file}.')
    if print_log: print(f'Terminal storage path: {device.paths.storage}.')
    if print_log: print(f'Terminal upload list path: {device.paths.upload_list}.')

    try:
        line = f'Terminal has {helper.get_free_space(cip, device.paths)} free bytes.'
    except:
        line = f'Failed to get free space on terminal.'
    device.log.append(line)
    if print_log: print(f'{line}')

    if not silent_mode:
        try:
            files = upload_med_list(cip, device)
            if redact_log: files = ['Redacted' for _ in files]
            line = f'Terminal has MED files: {files}.'
            if len(files) > 0: device.running_med_file = files[0]
        except:
            line = f'Failed to list MED files on terminal.'
        device.log.append(line)
        if print_log: print(f'{line}')

        try:
            files = upload_mer_list(cip, device)
            if redact_log: files = ['Redacted' for _ in files]
            line = f'Terminal has MER files: {files}.'
        except:
            line = f'Failed to list MER files on terminal.'
        device.log.append(line)
        if print_log: print(f'{line}')

    try:
        file = registry.get_startup_mer(cip)
        if redact_log: file = 'Redacted'
        line = f'Terminal startup file: {file}.'
        if file.lower().endswith('.mer'): device.startup_mer_file = file.split('\\')[-1]
    except:
        major_rev = int(device.me_version.split(".")[0])
        if major_rev <= 5:
            # For PanelView Plus 5.10 and earlier this registry key appears to be unavailable.
            line = f'Terminal startup file: could not be determined due to hardware version.'
        else:
            # If no startup app has been defined, it will also fail. 
            line = f'Terminal startup file: not configured.'
    device.log.append(line)
    if print_log: print(f'{line}')

def download_mer_file(cip: pycomm3.CIPDriver, device: types.MEDeviceInfo, file:types.MEFile, run_at_startup: bool, replace_comms: bool, delete_logs: bool) -> bool:
    # Create runtime folder
    try:
        helper.create_runtime_directory(cip, device.paths, file)
        device.log.append(f'Create runtime directory on terminal.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to create runtime directory on terminal.')
        return False

    # Get attributes
    #
    # Still no clue on what these are, or when/how they would change.
    # If they aren't changed by creating paths, could be moved ahead
    # to is_download_valid().
    try:
        get_unk1 = files.is_get_unk_valid(cip)
        if get_unk1:
            device.log.append(f'Got UNK1 attributes from terminal.')
        else:
            device.log.append(f'Got invalid UNK1 attributes from terminal.  Please file a bug report with all available information.')
            return False
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to get UNK1 attributes from terminal.')
        return False

    # Create a transfer instance on the terminal
    try:
        transfer_instance = files.create_transfer_instance_download(cip, file, f'{device.paths.storage}\\Rockwell Software\\RSViewME\\Runtime')
        device.log.append(f'Create transfer instance {transfer_instance} for download.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to create transfer instance for download')
        return False

    # Set attributes
    #
    # Still no clue what this is.  Might be setting file up for write?
    continue_download = True
    try:
        set_unk1 = files.is_set_unk_valid(cip)
        if set_unk1:
            device.log.append(f'Set UNK1 attributes on terminal.')
        else:
            device.log.append(f'Failed to set UNK1 attributes on terminal.  Please file a bug report with all available information.')
            continue_download = False
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to set UNK1 attributes on terminal.')
        continue_download = False

    # Transfer *.MER chunk by chunk
    try:
        if continue_download:
            files.download_mer(cip, transfer_instance, file.path)
            device.log.append(f'Downloaded {file.name} using transfer instance {transfer_instance}.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to download {file.name} using transfer instance {transfer_instance}.')

    # Delete transfer instance on the terminal
    try:
        files.delete_transfer_instance(cip, transfer_instance)
        device.log.append(f'Deleted transfer instance {transfer_instance}.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to delete transfer instance {transfer_instance}.')

    # Set *.MER to run at startup and then reboot
    try:
        if continue_download:
            if run_at_startup:
                helper.create_me_shortcut(cip, device.paths, file.name, replace_comms, delete_logs)
                device.log.append(f'Set file: {file.name} to run at startup with Replace Comms: {replace_comms}, Delete Logs: {delete_logs}.')
                reboot(cip, device)
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to set file to run at startup.')
        
    return continue_download

def upload_mer_file(cip: pycomm3.CIPDriver, device: types.MEDeviceInfo, file: types.MEFile, rem_file: types.MEFile) -> bool:
    # Verify file exists on terminal
    try:
        if helper.get_file_exists(cip, device.paths, rem_file):
            device.log.append(f'File {rem_file.name} exists on terminal.')
        else:
            device.log.append(f'File {rem_file.name} does not exist on terminal.')
            return False
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to check if file {rem_file.name} exists on terminal.')
        return False

    # Create a transfer instance on the terminal
    try:
        transfer_instance = files.create_transfer_instance_upload(cip, f'{device.paths.storage}\\Rockwell Software\\RSViewME\\Runtime\\{rem_file.name}')
        device.log.append(f'Create transfer instance {transfer_instance} for upload.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to create transfer instance for upload')
        return False

    # Transfer *.MER chunk by chunk
    try:
        files.upload_mer(cip, transfer_instance, file)
        device.log.append(f'Uploaded {rem_file.name} to {file.path} using transfer instance {transfer_instance}.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to upload {rem_file.name} to {file.path} using transfer instance {transfer_instance}.')

    # Delete transfer instance on the terminal
    try:
        files.delete_transfer_instance(cip, transfer_instance)
        device.log.append(f'Deleted transfer instance {transfer_instance}.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to delete transfer instance {transfer_instance}.')

    return True

def upload_med_list(cip: pycomm3.CIPDriver, device: types.MEDeviceInfo) -> list[str]:
    # Create list on the terminal
    try:
        helper.create_med_list(cip, device.paths)
        device.log.append(f'Created *.MED list on terminal.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to create *.MED list on terminal.')
        return None

    # Create transfer instance on the terminal
    try:
        transfer_instance = files.create_transfer_instance_upload(cip, device.paths.upload_list)
        device.log.append(f'Create transfer instance {transfer_instance} for upload.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to create transfer instance for upload')
        return None

    # Transfer list chunk by chunk
    file_list = None
    try:
        file_list = files.upload_list(cip, transfer_instance)
        device.log.append(f'Uploaded *.MED list using transfer instance {transfer_instance}.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to upload *.MED list using transfer instance {transfer_instance}.')

    # Delete transfer instance on the terminal
    try:
        files.delete_transfer_instance(cip, transfer_instance)
        device.log.append(f'Deleted transfer instance {transfer_instance}.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to delete transfer instance {transfer_instance}.')

    # Delete list on the terminal
    try:
        helper.delete_file_list(cip, device.paths)
        device.log.append(f'Deleted *.MED list on terminal.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to delete *.MED list on terminal.')

    return file_list

def upload_mer_list(cip: pycomm3.CIPDriver, device: types.MEDeviceInfo) -> list[str]:
    # Create *.MER list
    try:
        helper.create_mer_list(cip, device.paths)
        device.log.append(f'Created *.MER list on terminal.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to create *.MER list on terminal.')
        return None

    # Create transfer instance on the terminal
    try:
        transfer_instance = files.create_transfer_instance_upload(cip, device.paths.upload_list)
        device.log.append(f'Create transfer instance {transfer_instance} for upload.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to create transfer instance for upload')
        return None

    # Transfer list chunk by chunk
    file_list = None
    try:
        file_list = files.upload_list(cip, transfer_instance)
        device.files = file_list
        device.log.append(f'Uploaded *.MER list using transfer instance {transfer_instance}.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to upload *.MED list using transfer instance {transfer_instance}.')

    # Delete transfer instance on the terminal
    try:
        files.delete_transfer_instance(cip, transfer_instance)
        device.log.append(f'Deleted transfer instance {transfer_instance}.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to delete transfer instance {transfer_instance}.')

    # Delete list on the terminal
    try:
        helper.delete_file_list(cip, device.paths)
        device.log.append(f'Deleted *.MED list on terminal.')
    except Exception as e:
        device.log.append(f'Exception: {str(e)}')
        device.log.append(f'Failed to delete *.MED list on terminal.')

    return file_list

def reboot(cip: pycomm3.CIPDriver, device: types.MEDeviceInfo):
    cip1 = pycomm3.CIPDriver(cip._cip_path)
    cip1._cfg['socket_timeout'] = 0.25
    cip1.open()
    try:
        # Execute reboot
        device.log.append(f'Rebooting terminal.')
        helper.reboot(cip1, device.paths)

        # If we made it here... the reboot function didn't throw
        # an exception, which means it didn't reboot.
        #
        # Further investigation needed.
        device.log.append(f'Initial reboot unsuccessful.')

        # Currently tailoring additional action to terminals where:
        #
        # Firmware version is 12.108+
        # Application is set to load at startup
        # ME Station is set to run application at startup
        # Current App is defined
        if ((device.version_major < 12) or ((device.version_major == 12) and (device.version_minor < 108))):
            device.log.append(f'Did not attempt additional reboot because terminal is below minimum applicable version.')
            return
        if (registry.get_startup_options(cip) != 1):
            device.log.append(f'Did not attempt additional reboot because terminal Startup Options are not set to Run Current Application.')
            return
        startup_file = os.path.basename(registry.get_startup_mer(cip).replace('\\','/'))
        if not(startup_file.lower().endswith('.mer')):
            device.log.append(f'Did not attempt additional reboot because terminal does not have valid startup *.MER defined.')
            return

        # Rebuild existing ME Startup Shortcut on terminal
        delete_logs = registry.get_startup_delete_logs(cip)
        replace_comms = registry.get_startup_replace_comms(cip)
        device.log.append(f'Setting file: {startup_file} to run at startup with Replace Comms: {replace_comms}, Delete Logs: {delete_logs}.')
        helper.create_me_shortcut(cip, device.paths, startup_file, replace_comms, delete_logs)

        # Execute reboot
        device.log.append(f'Rebooting terminal.')
        helper.reboot(cip1, device.paths)
    except pycomm3.PycommError as e:
        # Unlike most CIP messages, this one is expected to
        # create an exception.  When it is received by the terminal,
        # the device reboots and breaks the socket.
        if (str(e) != 'failed to receive reply'): raise e
    
    cip1.close()


import os
from dataclasses import dataclass
from enum import StrEnum

@dataclass
class MEFile:
    name: str
    overwrite_requested: bool
    overwrite_required: bool
    path: str

    def get_ext(self) -> str:
        filename, extension = os.path.splitext(self.path)
        return extension.lower()

    def get_size(self) -> int:
        return os.path.getsize(self.path)
    
@dataclass
class MEDevicePaths:
    helper_file: str
    storage: str
    upload_list: str

@dataclass
class MEDeviceInfo:
    comms_path: str
    helper_version: str
    me_version: str
    version_major: int
    version_minor: int
    product_code: int
    product_name: str
    product_type: int
    log: list[str]
    files: list[str]
    running_med_file: str
    startup_mer_file: str
    paths: MEDevicePaths

@dataclass
class MEResponse(object):
    device: MEDeviceInfo
    status: str

class MEResponseStatus(StrEnum):
    SUCCESS = 'Success'
    FAILURE = 'Failure'


import pycomm3

from enum import StrEnum
from warnings import warn

from .. import messages
from .. import types 


# Known static value for successful creation of a folder.
# Further investigation needed.
CREATE_DIR_SUCCESS = 183 # 'b\xb7'

# Known functions available from RemoteHelper.
class HelperFunctions(StrEnum):
    REBOOT = 'BootTerminal'
    # ex: 'BootTerminal',''
    #       Returns a comms error since upon successful execution, terminal is reset.

    CREATE_DIRECTORY = 'CreateRemDirectory'
    # ex: 'CreateRemDirectory','{Path}'
    #       Returns a static value if successful, further investigation needed.

    CREATE_SHORTCUT = 'CreateRemShortcut'

    CREATE_ME_SHORTCUT = 'CreateRemMEStartupShortcut'
    # ex: 'CreateRemMEStartupShortcut','{Folder Path}:{Filename.MER}: /r /delay'
    #       Can also specify /o to replace comms and /d to delete logs at startup
    #       Returns 1 if shortcut created successfully, 0 otherwise.

    DELETE_DIRECTORY = 'DeleteRemDirectory'

    DELETE_FILE = 'DeleteRemFile'

    CREATE_FILE_LIST = 'FileBrowse'
    # ex: 'FileBrowse','{Search Path}\\*.{Search Extension}::{Results File Path}
    #       Returns 1 if results file generated successfully, 0 otherwise.

    GET_FILE_EXISTS = 'FileExists'
    # ex: 'FileExists','{File Path}'
    #       Returns 1 if file exists, 0 otherwise.

    GET_FILE_SIZE = 'FileSize'
    # ex: 'FileSize','{File Path}'
    #       Returns the number of bytes.

    CREATE_FOLDER_LIST = 'FolderBrowse'

    GET_FREE_SPACE = 'FreeSpace'
    # ex: 'FreeSpace','{Path}'
    #       Returns the number of bytes.
    GET_FILE_VERSION = 'GetFileVersion'
    GET_VERSION = 'GetVersion'
    # ex: 'GetVersion','{File Path}'
    #       Returns the a version string.

    START_PROCESS = 'InvokeEXE'
    GET_EXE_RUNNING = 'IsExeRunning'
    GET_FOLDER_EXISTS = 'StorageExists'
    # ex: 'StorageExists'.'{Path}'
    #       Returns 1 if folder exists, 0 otherwise.
    STOP_PROCESS = 'TerminateEXE'

def run_function(cip: pycomm3.CIPDriver, req_args):
    """
    Executes a function on the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal
        req_args: The request to execute.  Typically a list like ['DLL Path','Function Name','Function Args'] 

    Returns:
        resp_code (int): The message response code.
        resp_data (str): The function response data.

    Raises:
        Exception: If the message response code is not zero, indicating an error.

    Request Format:
        The request consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->N-1  | Request Arguments                                   |
        | Byte N        | Null footer                                         |

    Response Format:
        The response consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->3    | Response code (typically 0 = ran, other = error)    |
        | Bytes 4->N-1  | Function response                                   |
        | Bytes N       | Null footer                                         |

    Note:
        Only a certain subset of functions are whitelisted for access
        by this method.  They are documented in the HelperFunctions enum.
    """
    req_data = b''.join(arg.encode() + b'\x00' for arg in req_args)

    resp = messages.run_function(cip, req_data)
    if not resp: raise Exception(f'Failed to run function: {req_args}.')
    resp_code = int.from_bytes(resp.value[:4], byteorder='little', signed=False)
    resp_data = resp.value[4:].decode('utf-8').strip('\x00')
    return resp_code, resp_data

def create_directory(cip: pycomm3.CIPDriver, paths: types.MEDevicePaths, dir: str) -> bool:
    req_args = [paths.helper_file, HelperFunctions.CREATE_DIRECTORY, dir]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != CREATE_DIR_SUCCESS): raise Exception('Failed to create directory on terminal.')    
    return True

def create_me_shortcut(cip: pycomm3.CIPDriver, paths: types.MEDevicePaths, file: str, replace_comms: bool, delete_logs: bool) -> bool:
    """
    Setup a specific *.MER file to run at terminal startup.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal
        file (str): The *.MER file to use
        replace_comms (bool): If true, replaces the terminal comms setup with the configuration from the *.MER file.
        delete_logs (bool): If true, deletes logs at startup.

    Returns:
        Status (bool): True if function runs successfully.

    Request Format:
        {storage path}:{filename}: /r /delay /o /d

    Note:
        Even though the *.MER file resides in a subdirectory (ex: {storage path}\\Rockwell Software\\RSViewME\\Runtime\\{{filename}),
        it is only transmitted as {storage path}:{filename}.  The rest of the path is inserted by the RemoteHelper on the terminal.
    """ 
    replace_comms_param = ' /o' if replace_comms else ''
    delete_logs_param = ' /d' if delete_logs else ''
    shortcut = f'{paths.storage}:{file}: /r /delay{replace_comms_param}{delete_logs_param}'
    req_args = [paths.helper_file, HelperFunctions.CREATE_ME_SHORTCUT, shortcut]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to create ME Startup Shortcut on terminal: {file}, response code: {resp_code}, response data: {resp_data}.')
    return True

def create_med_list(cip:pycomm3.CIPDriver, paths: types.MEDevicePaths):
    req_args = [paths.helper_file,HelperFunctions.CREATE_FILE_LIST, f'\\Temp\\~MER.00\\*.med::{paths.upload_list}']
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Response code was not zero.  Examine packets.')
    return True

def create_mer_list(cip: pycomm3.CIPDriver, paths: types.MEDevicePaths):
    req_args = [paths.helper_file,HelperFunctions.CREATE_FILE_LIST, f'{paths.storage}\\Rockwell Software\\RSViewME\\Runtime\\*.mer::{paths.upload_list}']
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Response code was not zero.  Examine packets.')
    return True

def create_runtime_directory(cip: pycomm3.CIPDriver, paths: types.MEDevicePaths, file: types.MEFile) -> bool:
    # Create paths
    if not(create_directory(cip, paths, f'{paths.storage}')): return False
    if not(create_directory(cip, paths, f'{paths.storage}\\Rockwell Software')): return False
    if not(create_directory(cip, paths, f'{paths.storage}\\Rockwell Software\\RSViewME')): return False
    if not(create_directory(cip, paths, f'{paths.storage}\\Rockwell Software\\RSViewME\\Runtime')): return False
    return True

def delete_file(cip: pycomm3.CIPDriver, paths: types.MEDevicePaths, file: str) -> bool:
    req_args = [paths.helper_file,HelperFunctions.DELETE_FILE,file]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Failed to delete file on terminal: {file}, response code: {resp_code}, response data: {resp_data}.')
    return True

def delete_file_list(cip: pycomm3.CIPDriver, paths: types.MEDevicePaths) -> bool:
    return delete_file(cip, paths, paths.upload_list)

def get_file_exists(cip: pycomm3.CIPDriver, paths: types.MEDevicePaths, file: types.MEFile) -> bool:
    req_args = [paths.helper_file, HelperFunctions.GET_FILE_EXISTS, f'{paths.storage}\\Rockwell Software\\RSViewME\\Runtime\\{file.name}']
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): return False    
    return bool(int(resp_data))

def get_file_size(cip: pycomm3.CIPDriver, paths: types.MEDevicePaths, file: types.MEFile) -> int:
    req_args = [paths.helper_file, HelperFunctions.GET_FILE_SIZE, f'{paths.storage}\\Rockwell Software\\RSViewME\\Runtime\\{file.name}']
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Response code was not zero.  Examine packets.')
    return int(resp_data)

def get_folder_exists(cip: pycomm3.CIPDriver, paths: types.MEDevicePaths) -> bool:
    req_args = [paths.helper_file, HelperFunctions.GET_FOLDER_EXISTS, paths.storage]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0):
        warn(f'Response code was not zero.  Examine packets.')
        return False
    return bool(int(resp_data))

def get_free_space(cip: pycomm3.CIPDriver, paths: types.MEDevicePaths) -> int:
    req_args = [paths.helper_file, HelperFunctions.GET_FREE_SPACE, f'{paths.storage}\\Rockwell Software\\RSViewME\\Runtime\\']
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Response code was not zero.  Examine packets.')
    return int(resp_data)

def get_helper_version(cip: pycomm3.CIPDriver, paths: types.MEDevicePaths) -> str:
    req_args = [paths.helper_file, HelperFunctions.GET_VERSION, paths.helper_file]
    resp_code, resp_data = run_function(cip, req_args)
    if (resp_code != 0): raise Exception(f'Response code was not zero.  Examine packets.')
    return str(resp_data)

def reboot(cip: pycomm3.CIPDriver, paths: types.MEDevicePaths):
    req_args = [paths.helper_file, HelperFunctions.REBOOT,'']
    req_data = b''.join(arg.encode() + b'\x00' for arg in req_args)
    resp = messages.run_function(cip, req_data)



import pycomm3
import struct

from enum import IntEnum
from warnings import warn

from .. import messages
from .. import types

# When files are transferred, this is the maximum number of bytes
# used per message.  Quick tests up to 2000 bytes did succeed, >2000 bytes failed.
CHUNK_SIZE = 1984

# Known values associated with some file services, with unclear purpose or meaning.
# Further investigation needed.
GET_UNK1_VALUES = {
    b'\x02\x00'
}

# Known values associated with some file services, with unclear purpose or meaning.
# Further investigation needed.
GET_UNK2_VALUES = {
    b'\x01\x60',
    b'\x03\x41'
}

# Known values associated with some file services, with unclear purpose or meaning.
# Further investigation needed.
GET_UNK3_VALUES = {
    b'\x64\x00'
}

class TransferType(IntEnum):
    DOWNLOAD = int.from_bytes(b'\x01', byteorder='big')
    UPLOAD = int.from_bytes(b'\x00', byteorder='big')

def create_transfer_instance_download(cip: pycomm3.CIPDriver, file: types.MEFile, remote_path: str) -> int:
    """
    Creates a transfer instance for downloading from the local device to the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal
        file (MEFile): Identifying metadata for the file to be downloaded.
        remote_path (str): The remote path on the terminal where the file 
        will be downloaded to.

    Returns:
        int: The transfer instance returned from the device, that can be
        used for subsequent file transfer operations.

    Raises:
        Exception: If the transfer instance creation fails or if the response 
        contains unexpected values, indicating potential issues with the 
        transfer.

    Request Format:
        The request consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Byte 0        | Transfer Type (always 1 for file download)          |
        | Byte 1        | Overwrite flag (0 = New File, 1 = Overwrite)        |
        | Bytes 2->3    | Chunk size in bytes                                 |
        | Bytes 4->7    | File size in bytes                                  |
        | Bytes 8->N-1  | File name                                           |
        | Byte N        | Null footer                                         |

    Response Format:
        The response consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->1    | Message instance (should match request, 0x00)       |
        | Bytes 2->3    | Unknown purpose                                     |
        | Bytes 4->5    | Transfer instance (use this instance for download)  |
        | Bytes 6->7    | Chunk size in bytes                                 |

    Note:
        If the response message instance or unknown bytes are not zero, it 
        may indicate an incomplete transfer happened previously. In such cases,
        reboot the terminal and try again.
    """
    req_header = struct.pack('<BBHI', TransferType.DOWNLOAD, int(file.overwrite_required), CHUNK_SIZE, file.get_size())
    req_args = [f'{remote_path}\\{file.name}']
    req_data = req_header + b''.join(arg.encode() + b'\x00' for arg in req_args)

    resp = messages.create_transfer_instance(cip, req_data)
    resp_exception_text = 'Failed to create transfer instance for download'
    if not resp: raise Exception(f'{resp_exception_text}.  No message response.')
 
    resp_msg_instance, resp_unk1, resp_transfer_instance, resp_chunk_size = struct.unpack('<HHHH', resp.value)
    if (resp_msg_instance != 0): raise Exception(f'{resp_exception_text}.  Response message instance: {resp_msg_instance}, expected: 0.  There may be an incomplete transfer.  Reboot terminal and try again.')
    if (resp_unk1 != 0 ): raise Exception(f'{resp_exception_text}.  Response UNK1 bytes: {resp_unk1}, expected: 0.  Please file a bug report with all available information.')
    if (resp_chunk_size != CHUNK_SIZE): raise Exception(f'{resp_exception_text}.  Response chunk size: {resp_chunk_size}, expected: {CHUNK_SIZE}.  Please file a bug report with all available information.')
    return resp_transfer_instance

def create_transfer_instance_upload(cip: pycomm3.CIPDriver, remote_path: str) -> int:
    """
    Creates a transfer instance for uploading from the remote terminal to the local device.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal
        remote_path (str): The remote path on the terminal where the file 
        will be uploaded from.

    Returns:
        int: The transfer instance returned from the device, that can be
        used for subsequent file transfer operations.

    Raises:
        Exception: If the transfer instance creation fails or if the response 
        contains unexpected values, indicating potential issues with the 
        transfer.

    Request Format:
        The request consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Byte 0        | Transfer Type (always 0 for file upload)            |
        | Byte 1        | Unknown purpose (overwrite in download, N/A?)       |
        | Bytes 2->3    | Chunk size in bytes                                 |
        | Bytes 4->N-1  | File name                                           |
        | Byte N        | Null footer                                         |

    Response Format:
        The response consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->1    | Message instance (should match request, 0x00)       |
        | Bytes 2->3    | Unknown purpose                                     |
        | Bytes 4->5    | Transfer instance (use this instance for upload)    |
        | Bytes 6->7    | Chunk size in bytes                                 |
        | Bytes 8->11   | File size in bytes                                  |

    Note:
        If the response message instance or unknown bytes are not zero, it 
        may indicate an incomplete transfer happened previously. In such cases,
        reboot the terminal and try again.
    """
    req_header = struct.pack('<BBH', TransferType.UPLOAD, 0x00, CHUNK_SIZE)
    req_args = [f'{remote_path}']
    req_data = req_header + b''.join(arg.encode() + b'\x00' for arg in req_args)

    resp = messages.create_transfer_instance(cip, req_data)
    resp_exception_text = 'Failed to create transfer instance for upload'
    if not resp: raise Exception(f'{resp_exception_text}.  No message response.')

    resp_msg_instance, resp_unk1, resp_transfer_instance, resp_chunk_size, resp_file_size = struct.unpack('<HHHHI', resp.value)
    if (resp_msg_instance != 0): raise Exception(f'{resp_exception_text}.  Response message instance: {resp_msg_instance}, expected: 0.  There may be an incomplete transfer.  Reboot terminal and try again.')
    if (resp_unk1 != 0 ): raise Exception(f'{resp_exception_text}.  Response UNK1 bytes: {resp_unk1}, expected: 0.  Please file a bug report with all available information.')
    if (resp_chunk_size != CHUNK_SIZE): raise Exception(f'{resp_exception_text}.  Response chunk size: {resp_chunk_size}, expected: {CHUNK_SIZE}.  Please file a bug report with all available information.')
    return resp_transfer_instance

def delete_transfer_instance(cip: pycomm3.CIPDriver, transfer_instance: int):
    return messages.delete_transfer_instance(cip, transfer_instance)

def download(cip: pycomm3.CIPDriver, transfer_instance: int, source_data: bytearray) -> bool:
    """
    Downloads a file from the local device to the remote terminal.
    The transfer happens by breaking the file down into one or more
    chunks, with each chunk being sent via a CIP message and reassembled
    on the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal
        transfer_instance (int): The previously created file transfer instance.
        source_data (bytearray): The binary data of the file to be transferred.

    Returns:
        bool: True when transfer is complete.

    Raises:
        Exception: If the file transfer fails or if the response contains
        unexpected values, indicating potential issues with the transfer.

    Request Format:
        The request consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->3    | Chunk number                                        |
        | Bytes 4->5    | Chunk size in bytes                                 |
        | Bytes 6->N    | Chunk data                                          |

    Response Format:
        The response consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->3    | Unknown purpose                                     |
        | Bytes 4->7    | Chunk number echo                                   |
        | Bytes 8->11   | Next chunk number expected                          |

    Note:
        If the response message instance or unknown bytes are not zero, it 
        may indicate an incomplete transfer happened previously. In such cases,
        reboot the terminal and try again.
    """
    req_chunk_number = 1
    req_offset = 0
    while req_offset < len(source_data):
        req_chunk = source_data[req_offset:req_offset + CHUNK_SIZE]
        req_header = struct.pack('<IH', req_chunk_number, len(req_chunk))
        req_next_chunk_number = req_chunk_number + 1
        req_data = req_header + req_chunk

        # End of file
        if not req_chunk: break

        resp = messages.write_file_chunk(cip, transfer_instance, req_data)
        if not resp: raise Exception(f'Failed to write chunk {req_chunk_number} to terminal.')
        resp_unk1, resp_chunk_number, resp_next_chunk_number = struct.unpack('<III', resp.value)
        if (resp_unk1 != 0 ): raise Exception(f'Response unknown bytes {resp_unk1} are not zero.  Examine packets.')
        if (resp_chunk_number != req_chunk_number ): raise Exception(f'Response chunk number {resp_chunk_number} did not match request chunk number {req_chunk_number}.')
        if (resp_next_chunk_number != req_next_chunk_number): raise Exception(f'Response next chunk number {resp_next_chunk_number} did not match expected next chunk number {req_next_chunk_number}.')

        # Continue to next chunk
        req_chunk_number += 1
        req_offset += len(req_chunk)

    # Close out file
    req_data = b'\x00\x00\x00\x00\x02\x00\xff\xff'
    resp = messages.write_file_chunk(cip, transfer_instance, req_data)
    return True

def download_mer(cip: pycomm3.CIPDriver, transfer_instance: int, file: str):
    with open(file, 'rb') as source_file:
        return download(cip, transfer_instance, bytearray(source_file.read()))

def upload(cip: pycomm3.CIPDriver, transfer_instance: int) -> bytearray:
    """
    Uploads a file from the remote terminal to the local device.
    The transfer happens by breaking the file down into one or more
    chunks, with each chunk being sent via a CIP message and reassembled
    on the local device.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal
        transfer_instance (int): The previously created file transfer instance.

    Returns:
        bytearray: The raw binary data uploaded from the remote terminal.

    Raises:
        Exception: If the file transfer fails or if the response contains
        unexpected values, indicating potential issues with the transfer.

    Request Format:
        The request consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->3    | Chunk number                                        |

    Response Format:
        The response consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->3    | Unknown purpose                                     |
        | Bytes 4->7    | Chunk number echo                                   |
        | Bytes 8->9    | Chunk size in bytes                                 |
        | Bytes 10->N   | Chunk data                                          |

    Note:
        If the response message instance or unknown bytes are not zero, it 
        may indicate an incomplete transfer happened previously. In such cases,
        reboot the terminal and try again.
    """
    req_chunk_number = 1
    resp_binary = bytearray()
    while True:
        req_data = struct.pack('<I', req_chunk_number)

        resp = messages.read_file_chunk(cip, transfer_instance, req_data)
        if not resp: raise Exception(f'Failed to read chunk {req_chunk_number} to terminal.')
        resp_unk1 = int.from_bytes(resp.value[:4], byteorder='little', signed=False)
        resp_chunk_number = int.from_bytes(resp.value[4:8], byteorder='little', signed=False)
        resp_chunk_size = int.from_bytes(resp.value[8:9], byteorder='little', signed=False)
        resp_data = bytes(resp.value[10:])

        if (resp_unk1 != 0 ): raise Exception(f'Response unknown bytes {resp_unk1} are not zero.  Examine packets.')
        if (resp_chunk_number != req_chunk_number) and (resp_chunk_number != 0): raise Exception(f'Response chunk number {resp_chunk_number} did not match request chunk number {req_chunk_number}.')

        # End of file
        if (resp_chunk_number == 0) and (resp_chunk_size == 2) and (resp_data == b'\xff\xff'):
            #print('End of file')
            break

        # Write to mer list
        resp_binary += resp_data

        # Continue to next chunk
        req_chunk_number += 1

    return resp_binary

def upload_mer(cip: pycomm3.CIPDriver, transfer_instance: int, file: types.MEFile):
    resp_binary = upload(cip, transfer_instance)
    with open(file.path, 'wb') as dest_file:
        dest_file.write(resp_binary)

def upload_list(cip: pycomm3.CIPDriver, transfer_instance: int):
    resp_binary = upload(cip, transfer_instance)
    resp_str = "".join([chr(b) for b in resp_binary if b != 0])
    resp_list = resp_str.split(':')
    return resp_list

def is_get_unk_valid(cip: pycomm3.CIPDriver) -> bool:
    # I don't know what any of these three attributes are for yet.
    # It may be checking that the file exchange is available.
    resp = messages.get_attr_unk(cip, b'\x30\x01')
    if not resp: return False
    if resp.value not in GET_UNK1_VALUES:
        warn(f'Invalid UNK1 value.  Please file a bug report with all available information.')
        return False

    resp = messages.get_attr_unk(cip, b'\x30\x08')
    if not resp: return False
    if resp.value not in GET_UNK2_VALUES:
        warn(f'Invalid UNK2 value.  Please file a bug report with all available information.')
        return False

    resp = messages.get_attr_unk(cip, b'\x30\x09')
    if not resp: return False
    if resp.value not in GET_UNK3_VALUES:
        warn(f'Invalid UNK3 value.  Please file a bug report with all available information.')
        return False

    return True

def is_set_unk_valid(cip: pycomm3.CIPDriver) -> bool:
    # I don't know what setting this attribute does yet.
    # It may be marking the file exchange as in use.
    #
    resp = messages.set_attr_unk(cip, b'\x30\x01\xff\xff')
    if not resp: return False

    return True


import pycomm3

from enum import StrEnum
from .. import messages

# Known registry keys on the terminal that should be whitelisted for read access through RemoteHelper.
class RegKeys(StrEnum):
    CIP_VERSION_MAJOR = 'HKEY_LOCAL_MACHINE\SOFTWARE\Rockwell Software\RSLinxNG\CIP Identity\MajorRevision'                 # ex: 11
    CIP_VERSION_MINOR = 'HKEY_LOCAL_MACHINE\SOFTWARE\Rockwell Software\RSLinxNG\CIP Identity\MinorRevision'                 # ex: 1
    CIP_PRODUCT_CODE = 'HKEY_LOCAL_MACHINE\SOFTWARE\Rockwell Software\RSLinxNG\CIP Identity\ProductCode'                    # ex: 51
    CIP_PRODUCT_NAME = 'HKEY_LOCAL_MACHINE\SOFTWARE\Rockwell Software\RSLinxNG\CIP Identity\ProductName'                    # ex: PanelView Plus_6 1500
    CIP_PRODUCT_TYPE = 'HKEY_LOCAL_MACHINE\SOFTWARE\Rockwell Software\RSLinxNG\CIP Identity\ProductType'                    # ex: 24
    CIP_SERIAL_NUMBER = 'HKEY_LOCAL_MACHINE\SOFTWARE\Rockwell Software\RSLinxNG\CIP Identity\SerialNumber'                  # ex: 1234567
    CIP_VENDOR = 'HKEY_LOCAL_MACHINE\SOFTWARE\Rockwell Software\RSLinxNG\CIP Identity\Vendor'                               # ex: 1
    ME_VERSION = 'HKEY_LOCAL_MACHINE\SOFTWARE\Rockwell Software\RSView Enterprise\MEVersion'                                # ex: 11.00.25.230
    ME_STARTUP_APP = 'HKEY_LOCAL_MACHINE\SOFTWARE\Rockwell Software\RSViewME\Startup Options\CurrentApp'                    # ex: \Application Data\Rockwell Software\RSViewME\Runtime\{FileName}.mer
    ME_STARTUP_DELETE_LOGS = 'HKEY_LOCAL_MACHINE\SOFTWARE\Rockwell Software\RSViewME\Startup Options\DeleteLogFiles'        # ex: 0
    ME_STARTUP_LOAD_CURRENT = 'HKEY_LOCAL_MACHINE\SOFTWARE\Rockwell Software\RSViewME\Startup Options\LoadCurrentApp'       # ex: 1
    ME_STARTUP_REPLACE_COMMS = 'HKEY_LOCAL_MACHINE\SOFTWARE\Rockwell Software\RSViewME\Startup Options\ReplaceCommSettings' # ex: 0
    ME_STARTUP_OPTIONS = 'HKEY_LOCAL_MACHINE\SOFTWARE\Rockwell Software\RSViewME\Startup Options\StartupOptionsConfig'      # ex: 1

def get_value(cip: pycomm3.CIPDriver, key: str) -> str:
    """
    Gets a registry key's value from the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal
        key (str): The registry key to read.

    Returns:
        str: The registry value.

    Raises:
        Exception: If the message response code is not zero, indicating an error.

    Request Format:
        The request consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->N-1  | Registry key name                                   |
        | Byte N        | Null footer                                         |

    Response Format:
        The response consists of the following byte structure:

        | Byte Range    | Description                                         |
        |---------------|-----------------------------------------------------|
        | Bytes 0->3    | Response code (typically 0 = good, otherwise error) |
        | Bytes 4->7    | Unknown purpose                                     |
        | Bytes 8->N-1  | Registry key value                                  |
        | Bytes N       | Null footer                                         |

    Note:
        Only a certain subset of registry keys are whitelisted for access
        by this method.  They are documented in the RegKeys enum.
    """
    req_data = b''.join(arg.encode() + b'\x00' for arg in key)

    resp = messages.read_registry(cip, req_data)
    if not resp: raise Exception(f'Failed to read registry key: {key}')
    resp_code = int.from_bytes(resp.value[:4], byteorder='little', signed=False)
    if (resp_code != 0): raise Exception(f'Failed to read registry key: {key}, response code: {resp_code}.')

    resp_unk1 = int.from_bytes(resp.value[4:8], byteorder='little', signed=False)
    resp_value = str(resp.value[8:].decode('utf-8').strip('\x00'))
    return resp_value

def get_me_version(cip: pycomm3.CIPDriver) -> str:
    """
    Gets the ME Version on the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal

    Returns:
        str: ME Version string
    """
    return get_value(cip, [RegKeys.ME_VERSION])

def get_product_code(cip: pycomm3.CIPDriver) -> int:
    """
    Gets the product code on the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal

    Returns:
        int: Product Code value
    """
    return int(get_value(cip, [RegKeys.CIP_PRODUCT_CODE]))

def get_product_name(cip: pycomm3.CIPDriver) -> str:
    """
    Gets the product name on the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal

    Returns:
        str: Product Name value
    """
    return str(get_value(cip, [RegKeys.CIP_PRODUCT_NAME]))

def get_product_type(cip: pycomm3.CIPDriver) -> int:
    """
    Gets the product type on the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal

    Returns:
        int: Product Type value
    """
    return int(get_value(cip, [RegKeys.CIP_PRODUCT_TYPE]))

def get_startup_delete_logs(cip: pycomm3.CIPDriver) -> bool:
    """
    Gets the startup setting for Delete Logs on the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal

    Returns:
        bool: Don't (FALSE) or do (TRUE) delete log files at startup.
    """
    return bool(int(get_value(cip, [RegKeys.ME_STARTUP_DELETE_LOGS])))

def get_startup_load_current(cip: pycomm3.CIPDriver) -> bool:
    """
    Gets the startup setting for Load Current Application on the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal

    Returns:
        bool: Don't (FALSE) or do (TRUE) load the *.MER at startup.
    """
    return bool(int(get_value(cip, [RegKeys.ME_STARTUP_LOAD_CURRENT])))

def get_startup_mer(cip: pycomm3.CIPDriver) -> str:
    """
    Gets the *.MER file to load at startup on the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal

    Returns:
        str: Path to *.MER file.
    """
    return get_value(cip, [RegKeys.ME_STARTUP_APP])

def get_startup_options(cip: pycomm3.CIPDriver) -> int:
    """
    Gets the setting for On Startup on the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal

    Returns:
        int: At startup, 0 = Go to ME Station, 1 = Run Current Application,
        2 = Don't start ME Station      
    """
    return int(get_value(cip, [RegKeys.ME_STARTUP_OPTIONS]))

def get_startup_replace_comms(cip: pycomm3.CIPDriver) -> bool:
    """
    Gets the startup setting for Replace Communications on the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal

    Returns:
        bool: Don't (FALSE) or do (TRUE) replace terminal communications with *.MER
        communications settings at startup.
    """
    return bool(int(get_value(cip, [RegKeys.ME_STARTUP_REPLACE_COMMS])))

def get_version_major(cip: pycomm3.CIPDriver) -> int:
    """
    Gets the CIP Major Version from the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal

    Returns:
        int: The major version.

    Note:
        May not always align with the ME Version / Helper Version.
        For example, a v5.10 PanelView returned version 3.x.
    """
    return int(get_value(cip, [RegKeys.CIP_VERSION_MAJOR]))

def get_version_minor(cip: pycomm3.CIPDriver) -> int:
    """
    Gets the CIP Minor Version from the remote terminal.

    Args:
        cip (pycomm3.CIPDriver): CIPDriver to communicate with the terminal

    Returns:
        int: The minor version.

    Note:
        May not always align with the ME Version / Helper Version.
        For example, a v5.10 PanelView returned version 3.x.
    """
    return int(get_value(cip, [RegKeys.CIP_VERSION_MINOR]))
